# Copyright 2015-2017, Truveris Inc. All Rights Reserved.

"""
Extract and generate overpunch formatted numbers.
"""

from decimal import Decimal, ROUND_HALF_UP


__copyright__ = "(c) 2015 Truveris"
__version__ = "1.1"


EXTRACT_REF = {
    "0": ("+", "0"),
    "1": ("+", "1"),
    "2": ("+", "2"),
    "3": ("+", "3"),
    "4": ("+", "4"),
    "5": ("+", "5"),
    "6": ("+", "6"),
    "7": ("+", "7"),
    "8": ("+", "8"),
    "9": ("+", "9"),
    "{": ("+", "0"),
    "A": ("+", "1"),
    "B": ("+", "2"),
    "C": ("+", "3"),
    "D": ("+", "4"),
    "E": ("+", "5"),
    "F": ("+", "6"),
    "G": ("+", "7"),
    "H": ("+", "8"),
    "I": ("+", "9"),
    "}": ("-", "0"),
    "J": ("-", "1"),
    "K": ("-", "2"),
    "L": ("-", "3"),
    "M": ("-", "4"),
    "N": ("-", "5"),
    "O": ("-", "6"),
    "P": ("-", "7"),
    "Q": ("-", "8"),
    "R": ("-", "9"),
}


def extract(raw, decimals=2):
    """Extract a number in the overpunch format to a Decimal object.

    :param raw: The formatted value.
    :param decimals: The implied decimal precision of this number (default: 2).

    """

    length = len(raw)
    last_char = raw[length - 1]
    (sign, cent) = EXTRACT_REF[last_char]

    if not decimals:
        core = raw[:-1]
    else:
        core = raw[:length-decimals] + "." + raw[length-decimals:-1]

    return Decimal(sign + core + cent)


FORMAT_REF = {
    ("+", "0"): "{",
    ("+", "1"): "A",
    ("+", "2"): "B",
    ("+", "3"): "C",
    ("+", "4"): "D",
    ("+", "5"): "E",
    ("+", "6"): "F",
    ("+", "7"): "G",
    ("+", "8"): "H",
    ("+", "9"): "I",
    ("-", "0"): "}",
    ("-", "1"): "J",
    ("-", "2"): "K",
    ("-", "3"): "L",
    ("-", "4"): "M",
    ("-", "5"): "N",
    ("-", "6"): "O",
    ("-", "7"): "P",
    ("-", "8"): "Q",
    ("-", "9"): "R",
}


def format(val, decimals=2, rounding=ROUND_HALF_UP):
    """Convert a number to an overpunch-formatted string.

    :param val: The number to format.
    :param decimals: How many decimals are implied in the formatted value
        (default: 2).
    :param rounding: When rounding a value during quantization, how that value
        is to be rounded. Appropriate values are the rounding constants from
        the ``decimal`` library (default: ROUND_HALF_UP).

    """

    if not isinstance(val, Decimal):
        val = Decimal(str(val))

    if val.is_nan():
        raise ValueError("{} is NaN".format(val))

    # force the correct number of decimal digits
    quantize_str = "1"
    if decimals:
        quantize_str = quantize_str + "." + ("0" * decimals)

    val = val.quantize(Decimal(quantize_str), rounding)

    # we'll need this to figure out how to format the last character
    if val.is_signed():
        sign = "-"
    else:
        sign = "+"

    # get rid of the sign
    val = abs(val)

    # turn into a string
    val = str(val)

    # remove the "."
    val = val.replace(".", "")

    # split, replace the last digit with the right format character, join
    parts = list(val)
    parts[-1] = FORMAT_REF[(sign, parts[-1])]

    return "".join(parts)



# Copyright 2015-2017, Truveris Inc. All Rights Reserved.

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

from overpunch import __version__


setup(
    name="overpunch",
    version=__version__,
    description="Overpunch Parser/Formatter",
    author="Truveris Inc.",
    author_email="dev@truveris.com",
    url="http://github.com/truveris/overpunch",
    test_suite="nose.collector",
    packages=find_packages(exclude=["ez_setup"]),
    classifiers=[
        "Intended Audience :: Healthcare Industry",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Text Processing",
    ],
)




# Copyright 2015-2017, Truveris Inc. All Rights Reserved.

import sys
import random
import unittest
from decimal import Decimal, ROUND_FLOOR

import overpunch


class TestCase(unittest.TestCase):

    """
    Test cases for overpunched extraction and formatting.
    """

    def test_extract_no_decimal_positive(self):
        d = Decimal("1234.50")
        self.assertEquals(overpunch.extract("12345{"), d)

    def test_extract_no_decimal_negative(self):
        d = Decimal("-1234.50")
        self.assertEquals(overpunch.extract("12345}"), d)

    def test_extract_4_decimal_negative(self):
        d = Decimal("-12.3450")
        self.assertEquals(overpunch.extract("12345}", decimals=4), d)

    def test_extract_0_decimal_negative(self):
        d = Decimal("-123450")
        self.assertEquals(overpunch.extract("12345}", decimals=0), d)

    def test_format_no_decimal_positive(self):
        d = Decimal("1234.50")
        self.assertEquals("12345{", overpunch.format(d))

    def test_format_no_decimal_negative(self):
        d = Decimal("-1234.50")
        self.assertEquals("12345}", overpunch.format(d))

    def test_format_4_decimal_negative(self):
        d = Decimal("-12.3450")
        self.assertEquals("12345}", overpunch.format(d, decimals=4))

    def test_format_0_decimal_negative(self):
        d = Decimal("-123450")
        self.assertEquals("12345}", overpunch.format(d, decimals=0))

    def test_format_2_decimal_round_default(self):
        d = Decimal("12.3450")
        self.assertEquals("123E", overpunch.format(d, decimals=2))

    def test_format_2_decimal_negative_round_default(self):
        d = Decimal("-12.3450")
        self.assertEquals("123N", overpunch.format(d, decimals=2))

    def test_format_2_decimal_round_custom(self):
        d = Decimal("12.3450")
        self.assertEquals("123D", overpunch.format(d, decimals=2,
                                                   rounding=ROUND_FLOOR))

    def test_format_2_decimal_negative_round_custom(self):
        d = Decimal("-12.3450")
        self.assertEquals("123N", overpunch.format(d, decimals=2,
                                                   rounding=ROUND_FLOOR))

    def test_nan_raises(self):
        d = Decimal("NaN")
        with self.assertRaises(ValueError):
            overpunch.format(d)

    def test_format_integer(self):
        d = 150
        self.assertEquals("15{", overpunch.format(d, decimals=0))

    def test_random(self):
        for i in range(1000):
            d = Decimal(random.randint(0, sys.maxsize)) / 100
            if random.random() > 0.5:
                d = -d

            self.assertEquals(overpunch.extract(overpunch.format(d)), d)
